#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <windows.h>
#include <GL/gl.h>

#include "objload.h"

static void cacheobjtrinormal(OBJECT *obj)
{
	int i, j;

	for(i = 0; i < obj->cobjs; i++)
	{
		obj->objs[i].tnor = malloc(sizeof(float) * 3 * obj->objs[i].ctris);

		for(j = 0; j < obj->objs[i].ctris; j++)
		{
			float v1[3], v2[3], v3[3], n;

			v1[0] = (obj->objs[i].vtxs[obj->objs[i].tris[j][1]][0] - obj->objs[i].vtxs[obj->objs[i].tris[j][0]][0]);
			v1[1] = (obj->objs[i].vtxs[obj->objs[i].tris[j][1]][1] - obj->objs[i].vtxs[obj->objs[i].tris[j][0]][1]);
			v1[2] = (obj->objs[i].vtxs[obj->objs[i].tris[j][1]][2] - obj->objs[i].vtxs[obj->objs[i].tris[j][0]][2]);
			v2[0] = (obj->objs[i].vtxs[obj->objs[i].tris[j][2]][0] - obj->objs[i].vtxs[obj->objs[i].tris[j][1]][0]);
			v2[1] = (obj->objs[i].vtxs[obj->objs[i].tris[j][2]][1] - obj->objs[i].vtxs[obj->objs[i].tris[j][1]][1]);
			v2[2] = (obj->objs[i].vtxs[obj->objs[i].tris[j][2]][2] - obj->objs[i].vtxs[obj->objs[i].tris[j][1]][2]);

			v3[0] = v1[1]*v2[2] - v1[2]*v2[1];
			v3[1] = v1[2]*v2[0] - v1[0]*v2[2];
			v3[2] = v1[0]*v2[1] - v1[1]*v2[0];

			n = (float)sqrt(v3[0]*v3[0] + v3[1]*v3[1] + v3[2]*v3[2]);
			obj->objs[i].tnor[j][0] = v3[0] / n;
			obj->objs[i].tnor[j][1] = v3[1] / n;
			obj->objs[i].tnor[j][2] = v3[2] / n;

			// glNormal3fv(v3);

			/*
			v2[0] = 0;
			v2[1] = 0;
			v2[2] = 0;
			{
				int k, l;

				for(k = 0; k < 3; k++)
					for(l = 0; l < 3; l++)
						v2[l] += obj->objs[i].vtxs[obj->objs[i].tris[j][k]][l];
				v2[0] /= 3;
				v2[1] /= 3;
				v2[2] /= 3;

				glBegin(GL_LINES);
				glVertex3f(v2[0]          , v2[1]          , v2[2]          );
				glVertex3f(v2[0]+v3[0]*.1f, v2[1]+v3[1]*.1f, v2[2]+v3[2]*.1f);
				glEnd();
			}
			//*/
		}
	}
}

static void cacheobjvtxnormal(OBJECT *obj)
{
	int i, j, k;

	for(i = 0; i < obj->cobjs; i++)
	{
		obj->objs[i].vnor = malloc(sizeof(float) * 3 * obj->objs[i].cvtxs);

		for(j = 0; j < obj->objs[i].cvtxs; j++)
		{
			obj->objs[i].vnor[j][0] = 0.0f;
			obj->objs[i].vnor[j][1] = 0.0f;
			obj->objs[i].vnor[j][2] = 0.0f;
		}

		for(j = 0; j < obj->objs[i].ctris; j++)
		{
			for(k = 0; k < 3; k++)
			{
				float v[3], n;
				int vidx = obj->objs[i].tris[j][k];

				v[0] = obj->objs[i].vnor[vidx][0] + obj->objs[i].tnor[j][0];
				v[1] = obj->objs[i].vnor[vidx][1] + obj->objs[i].tnor[j][1];
				v[2] = obj->objs[i].vnor[vidx][2] + obj->objs[i].tnor[j][2];

				n = (float)sqrt(v[0]*v[0] + v[1]*v[1] + v[2]*v[2]);
				obj->objs[i].vnor[vidx][0] = v[0] / n;
				obj->objs[i].vnor[vidx][1] = v[1] / n;
				obj->objs[i].vnor[vidx][2] = v[2] / n;
			}
		}
	}
}

struct tagDataReader{
	void *data;
	long offs;
	long size;
	size_t (*read)(void *buf, size_t size, struct tagDataReader *data);
	int (*seek)(struct tagDataReader *data, long offs, int type);
	int (*eof)(struct tagDataReader *data);
	int (*getc)(struct tagDataReader *data);
};

static OBJECT *_load3ds(struct tagDataReader *data)
{
	OBJECT *obj;

	obj = malloc(sizeof(*obj));
	obj->cobjs = 0;
	obj->cmats = 0;
	obj->objs = NULL;
	obj->mats = NULL;

	for(;;)
	{
		unsigned short chid;	// 2byte
		unsigned long size;		// 4byte

		data->read(&chid, sizeof(chid), data);
		if(data->eof(data))	// EOF indicator on when last "read" is EOF, not "now position".
			break;
		data->read(&size, sizeof(size), data);

		switch(chid)
		{
		case 0x4D4D:	// main chunk
			break;

		case 0x0002:	// version
			data->seek(data, +4, SEEK_CUR);
			break;

		//case 0x3D3D:	// object info
		//	break;

		case 0x3D3D:	// material container
			break;
		case 0xAFFF:	// material
			obj->cmats++;
			obj->mats = realloc(obj->mats, obj->cmats * sizeof(*obj->mats));
			break;
		case 0xA000:	// material name
			data->read(obj->mats[obj->cmats - 1].name, size-6, data);	// contains null
			break;
		case 0xA010:	// material ambient
			data->seek(data, +6, SEEK_CUR);	// RGBchunk
			data->read(obj->mats[obj->cmats - 1].amb, 3, data);
			break;
		case 0xA020:	// material diffuse
			data->seek(data, +6, SEEK_CUR);	// RGBchunk
			data->read(obj->mats[obj->cmats - 1].dif, 3, data);
			break;
		case 0xA030:	// material specular
			data->seek(data, +6, SEEK_CUR);	// RGBchunk
			data->read(obj->mats[obj->cmats - 1].spe, 3, data);
			break;

		case 0x4000:	// object
			{
				int i = 0;
				char str[17];

				for(;;)
				{
					char c = data->getc(data);

					str[i] = c;

					i++;
					if(!c)
						break;
				}

				obj->cobjs++;
				obj->objs = realloc(obj->objs, obj->cobjs * sizeof(*obj->objs));

				strcpy(obj->objs[obj->cobjs - 1].name, str);
				obj->objs[obj->cobjs - 1].cvtxs = 0;
				obj->objs[obj->cobjs - 1].ctris = 0;
				obj->objs[obj->cobjs - 1].vtxs = NULL;
				obj->objs[obj->cobjs - 1].vuvs = NULL;
				obj->objs[obj->cobjs - 1].tris = NULL;
				obj->objs[obj->cobjs - 1].tmat = NULL;
			}
			break;
		case 0x4100:	// mesh
			break;
		case 0x4110:	// mesh vertex
			{
				int i, j;

				data->read(&obj->objs[obj->cobjs - 1].cvtxs, sizeof(unsigned short), data);
				obj->objs[obj->cobjs - 1].vtxs = malloc(sizeof(float) * 3 * obj->objs[obj->cobjs - 1].cvtxs);
				for(i = 0; i < obj->objs[obj->cobjs - 1].cvtxs; i++)
					for(j = 0; j < 3; j++)
						data->read(&obj->objs[obj->cobjs - 1].vtxs[i][j], sizeof(float), data);
			}
			break;
		case 0x4120:	// mesh face
			{
				int i, j;

				data->read(&obj->objs[obj->cobjs - 1].ctris, sizeof(unsigned short), data);
				obj->objs[obj->cobjs - 1].tris = malloc(sizeof(unsigned short) * 3 * obj->objs[obj->cobjs - 1].ctris);
				obj->objs[obj->cobjs - 1].tmat = malloc(sizeof(unsigned short) * 3 * obj->objs[obj->cobjs - 1].ctris);
				for(i = 0; i < obj->objs[obj->cobjs - 1].ctris; i++)
				{
					for(j = 0; j < 3; j++)
						data->read(&obj->objs[obj->cobjs - 1].tris[i][j], sizeof(unsigned short), data);
					data->seek(data, +2, SEEK_CUR);
				}
			}
			break;
		case 0x4130:	// mesh material(in mesh face)
			{
				int i = 0;
				char str[256];

				for(;;)
				{
					char c = data->getc(data);

					str[i] = c;

					i++;
					if(!c)
						break;
				}

				// find material
				for(i = 0; i < obj->cmats; i++)
				{
					if(!strcmp(obj->mats[i].name, str))
						break;
				}

				if(!(i < obj->cmats))
				{	// NOT FOUND
					return NULL;
				}else
				{	// found
					unsigned short matfaces;
					int j;

					data->read(&matfaces, sizeof(matfaces), data);

					for(j = 0; j < matfaces; j++)
					{
						unsigned short idx;

						data->read(&idx, sizeof(idx), data);
						obj->objs[obj->cobjs - 1].tmat[idx] = i;
					}
				}
			}

			break;
		case 0x4140:	// mesh uv
			{
				int i, j;

				//fread(&obj->objs[obj->cobjs - 1].cvtxs, sizeof(unsigned short), 1, fp);
				data->seek(data, +2, SEEK_CUR);	// uv count, same as vertexs
				obj->objs[obj->cobjs - 1].vuvs = malloc(sizeof(float) * 2 * obj->objs[obj->cobjs - 1].cvtxs);
				for(i = 0; i < obj->objs[obj->cobjs - 1].cvtxs; i++)
					for(j = 0; j < 2; j++)
						data->read(&obj->objs[obj->cobjs - 1].vuvs[i][j], sizeof(float), data);
			}
			break;

		default:
			data->seek(data, size-6, SEEK_CUR);
			break;
		}
	}

	cacheobjtrinormal(obj);
	cacheobjvtxnormal(obj);

	return obj;
}

static size_t stdread(void *buf, size_t size, struct tagDataReader *data)
{
	return fread(buf, size, 1, data->data);
}
static int stdseek(struct tagDataReader *data, long offs, int type)
{
	return fseek(data->data, offs, type);
}
static int stdeof(struct tagDataReader *data)
{
	return feof((FILE*)data->data);
}
static int stdgetc(struct tagDataReader *data)
{
	return getc((FILE*)data->data);
}

OBJECT *load3ds(const char *filename)
{
	FILE *fp;
	struct tagDataReader data;
	OBJECT *res;

	if((fp = fopen(filename, "rb")) == NULL)
	{
		return NULL;
	}

	data.data = fp;
	data.read = stdread;
	data.seek = stdseek;
	data.eof = stdeof;
	data.getc = stdgetc;

	res = _load3ds(&data);

	fclose(fp);

	return res;
}

static size_t memread(void *buf, size_t size, struct tagDataReader *data)
{
	if(data->offs + (long)size > data->size)
		size = data->size - data->offs;

	if(size > 0)
		memcpy(buf, (char*)data->data + data->offs, size);

	data->offs += size;

	return size;
}
static int memseek(struct tagDataReader *data, long offs, int type)
{
	long dest;

	switch(type)
	{
	case SEEK_SET:	dest = offs;	break;
	case SEEK_CUR:	dest = data->offs + offs;	break;
	case SEEK_END:	dest = data->size;	break;
	}

	if(dest < 0)
		dest = 0;
	if(dest > data->size)
		dest = data->size;

	data->offs = dest;

	return 0;
}
static int memeof(struct tagDataReader *data)
{
	return data->offs == data->size;
}
static int memgetc(struct tagDataReader *data)
{
	char c;

	if(!data->read(&c, sizeof(char), data))
		return 0;

	return c;
}

OBJECT *load3dsfrommem(const char *buf, int size)
{
	struct tagDataReader data;
	OBJECT *res;

	data.data = (char*)buf;
	data.size = size;
	data.offs = 0;
	data.read = memread;
	data.seek = memseek;
	data.eof = memeof;
	data.getc = memgetc;

	res = _load3ds(&data);

	return res;
}
